﻿private import std.stdio;   // writefln
private import std.ctype;   // isdigit
private import std.stream;  // BufferedFile

alias char[] string;
alias string[] stringarray;

/// Strips non-digit characters from the string (COW)
string stripNonDigit( in string line ) 
{
    string ret;
    foreach(uint i, c; line) {
        // Error: Error: std.ctype.isdigit at C:\dmd\src\phobos\std\ctype.d(37) conflicts with std.stream.isdigit at C:\dmd\src\phobos\std\stream.d(2924)
	if (!std.ctype.isdigit(c)) {
	    if (!ret)
    		ret = line[0..i];	
	}	
	else if (ret)
	    ret ~= c;	
    }	
    return ret?ret:line;
}

unittest {
    assert( stripNonDigit("\'13-=2 4kop") == "1324" );
}

/// Converts a word into a number, ignoring all non alpha characters
string wordToNum( in string word )
{
version(UseStandardTranslation) {
const char[256] TRANSLATE =	// translation table for standard phones
    "                                "  // 0
    "                0123456789      "	// 32
    " 22233344455566677778889999     "  // 64
    " 22233344455566677778889999     "
    "                                "
    "                                "
    "                                "
    "                                ";
} else {
const char[256] TRANSLATE =	// translation table for the task at hand
    "                                "  // 0
    "                0123456789      "	// 32
    " 57630499617851881234762239     "  // 64
    " 57630499617851881234762239     "
    "                                "
    "                                "
    "                                "
    "                                ";
}//version(UseStandardTranslation)
    string ret;
    foreach(c; cast(ubyte[])word)
	if (TRANSLATE[c] != ' ')
	    ret ~= TRANSLATE[c];
    return ret;
}

unittest {
version(UseStandardTranslation) { } else {
    assert( wordToNum("E | J N Q | R W X | D S Y | F T | A M | C I V | B K U | L O P | G H Z") == "01112223334455666777888999" );
    assert( wordToNum("e | j n q | r w x | d s y | f t | a m | c i v | b k u | l o p | g h z") == "01112223334455666777888999" );
}
    assert( wordToNum("0 |   1   |   2   |   3   |  4  |  5  |   6   |   7   |   8   |   9") == "0123456789" );
}

void main( string[] args )
{
    stringarray[string]	num2words;   // maps a number to an array of words

    foreach(string word; new BufferedFile("dictionary.txt") )
	num2words[ wordToNum(word) ] ~= word.dup;	    // must dup

    /// Finds all alternatives for the given number (should have been stripped from non-digit characters)
    stringarray _FindWords( string numbers, bool digitok )
    in {
	assert(numbers.length > 0);	
    }	
    out(result) {
	foreach (a; result)
	    assert( wordToNum(a) == numbers );
    }	
    body {
	stringarray ret;
	for (uint t=1; t<=numbers.length; ++t) {
	    auto alternatives = numbers[0..t] in num2words;
	    if (!alternatives)
		continue;
	    if (numbers.length > t) {
		// Combine all current alternatives with all alternatives of the rest
   		foreach (a2; _FindWords( numbers[t..$], true ) )    // 1 digit is allowed
		    foreach(a1; *alternatives)
			ret ~= a1 ~ " " ~ a2;
	    }
	    else	
		ret ~= *alternatives;		// append these alternatives
	}
	// Try to keep 1 digit (only if we're allowed and no other alternatives were found)
	if (digitok && ret.length == 0) {
	    if(numbers.length > 1) {
		// Combine 1 digit with all altenatives from the rest
		foreach (a; _FindWords( numbers[1..$], false ) )    // next time, no digit
		    ret ~= numbers[0..1] ~ " " ~ a;
	    }	
	    else	
		ret ~= numbers[0..1];		// just append this digit
	}	
        return ret;
    }

    /// Finds all alternatives for the given phone number
    stringarray FindWords( string phone_number )
    {
	return _FindWords( stripNonDigit(phone_number), true );	
    }	
	
    if (args.length > 1) {
	// Convert the phone numbers provided on the command line instead
	foreach(string phone; args[1..$] )
	    foreach(alternative; FindWords( phone ) )
		writefln(phone, ": ", alternative );
    } 
    else {    
	// Read the phone numbers
	foreach(string phone; new BufferedFile("input.txt") )
	    foreach(alternative; FindWords( phone ) )
		writefln(phone, ": ", alternative );
    }
}
